# /*##########################################################################
#
# Copyright (c) 2017-2019 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""Tests for ImageAlphaSlider"""


__authors__ = ["P. Knobel"]
__license__ = "MIT"
__date__ = "28/03/2017"

import numpy
import unittest

from silx.gui import qt
from silx.gui.utils.testutils import TestCaseQt
from silx.gui.plot import PlotWidget
from silx.gui.plot import AlphaSlider


class TestActiveImageAlphaSlider(TestCaseQt):
    def setUp(self):
        super(TestActiveImageAlphaSlider, self).setUp()
        self.plot = PlotWidget()
        self.aslider = AlphaSlider.ActiveImageAlphaSlider(plot=self.plot)
        self.aslider.setOrientation(qt.Qt.Horizontal)

        toolbar = qt.QToolBar("plot", self.plot)
        toolbar.addWidget(self.aslider)
        self.plot.addToolBar(toolbar)

        self.plot.show()
        self.qWaitForWindowExposed(self.plot)

        self.mouseMove(self.plot)  # Move to center
        self.qapp.processEvents()

    def tearDown(self):
        self.qapp.processEvents()
        self.plot.setAttribute(qt.Qt.WA_DeleteOnClose)
        self.plot.close()
        del self.plot
        del self.aslider

        super(TestActiveImageAlphaSlider, self).tearDown()

    def testWidgetEnabled(self):
        # no active image initially, slider must be deactivate
        self.assertFalse(self.aslider.isEnabled())

        self.plot.addImage(numpy.array([[0, 1, 2], [3, 4, 5]]))
        # now we have an active image
        self.assertTrue(self.aslider.isEnabled())

        self.plot.setActiveImage(None)
        self.assertFalse(self.aslider.isEnabled())

    def testGetImage(self):
        self.plot.addImage(numpy.array([[0, 1, 2], [3, 4, 5]]))
        self.assertEqual(self.plot.getActiveImage(),
                         self.aslider.getItem())

        self.plot.addImage(numpy.array([[0, 1, 3], [2, 4, 6]]), legend="2")
        self.plot.setActiveImage("2")
        self.assertEqual(self.plot.getImage("2"),
                         self.aslider.getItem())

    def testGetAlpha(self):
        self.plot.addImage(numpy.array([[0, 1, 2], [3, 4, 5]]), legend="1")
        self.aslider.setValue(137)
        self.assertAlmostEqual(self.aslider.getAlpha(),
                               137. / 255)


class TestNamedImageAlphaSlider(TestCaseQt):
    def setUp(self):
        super(TestNamedImageAlphaSlider, self).setUp()
        self.plot = PlotWidget()
        self.aslider = AlphaSlider.NamedImageAlphaSlider(plot=self.plot)
        self.aslider.setOrientation(qt.Qt.Horizontal)

        toolbar = qt.QToolBar("plot", self.plot)
        toolbar.addWidget(self.aslider)
        self.plot.addToolBar(toolbar)

        self.plot.show()
        self.qWaitForWindowExposed(self.plot)

        self.mouseMove(self.plot)  # Move to center
        self.qapp.processEvents()

    def tearDown(self):
        self.qapp.processEvents()
        self.plot.setAttribute(qt.Qt.WA_DeleteOnClose)
        self.plot.close()
        del self.plot
        del self.aslider

        super(TestNamedImageAlphaSlider, self).tearDown()

    def testWidgetEnabled(self):
        # no image set initially, slider must be deactivate
        self.assertFalse(self.aslider.isEnabled())

        self.plot.addImage(numpy.array([[0, 1, 2], [3, 4, 5]]), legend="1")
        self.aslider.setLegend("1")
        # now we have an image set
        self.assertTrue(self.aslider.isEnabled())

    def testGetImage(self):
        self.plot.addImage(numpy.array([[0, 1, 2], [3, 4, 5]]), legend="1")
        self.plot.addImage(numpy.array([[0, 1, 3], [2, 4, 6]]), legend="2")
        self.aslider.setLegend("1")
        self.assertEqual(self.plot.getImage("1"),
                         self.aslider.getItem())

        self.aslider.setLegend("2")
        self.assertEqual(self.plot.getImage("2"),
                         self.aslider.getItem())

    def testGetAlpha(self):
        self.plot.addImage(numpy.array([[0, 1, 2], [3, 4, 5]]), legend="1")
        self.aslider.setLegend("1")
        self.aslider.setValue(128)
        self.assertAlmostEqual(self.aslider.getAlpha(),
                               128. / 255)


class TestNamedScatterAlphaSlider(TestCaseQt):
    def setUp(self):
        super(TestNamedScatterAlphaSlider, self).setUp()
        self.plot = PlotWidget()
        self.aslider = AlphaSlider.NamedScatterAlphaSlider(plot=self.plot)
        self.aslider.setOrientation(qt.Qt.Horizontal)

        toolbar = qt.QToolBar("plot", self.plot)
        toolbar.addWidget(self.aslider)
        self.plot.addToolBar(toolbar)

        self.plot.show()
        self.qWaitForWindowExposed(self.plot)

        self.mouseMove(self.plot)  # Move to center
        self.qapp.processEvents()

    def tearDown(self):
        self.qapp.processEvents()
        self.plot.setAttribute(qt.Qt.WA_DeleteOnClose)
        self.plot.close()
        del self.plot
        del self.aslider

        super(TestNamedScatterAlphaSlider, self).tearDown()

    def testWidgetEnabled(self):
        # no Scatter set initially, slider must be deactivate
        self.assertFalse(self.aslider.isEnabled())

        self.plot.addScatter([0, 1, 2], [2, 3, 4], [5, 6, 7],
                             legend="1")
        self.aslider.setLegend("1")
        # now we have an image set
        self.assertTrue(self.aslider.isEnabled())

    def testGetScatter(self):
        self.plot.addScatter([0, 1, 2], [2, 3, 4], [5, 6, 7],
                             legend="1")
        self.plot.addScatter([0, 10, 20], [20, 30, 40], [50, 60, 70],
                             legend="2")
        self.aslider.setLegend("1")
        self.assertEqual(self.plot.getScatter("1"),
                         self.aslider.getItem())

        self.aslider.setLegend("2")
        self.assertEqual(self.plot.getScatter("2"),
                         self.aslider.getItem())

    def testGetAlpha(self):
        self.plot.addScatter([0, 10, 20], [20, 30, 40], [50, 60, 70],
                             legend="1")
        self.aslider.setLegend("1")
        self.aslider.setValue(128)
        self.assertAlmostEqual(self.aslider.getAlpha(),
                               128. / 255)



# /*##########################################################################
#
# Copyright (c) 2015-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""This module provides a simple generic notification system."""

__authors__ = ["T. Vincent"]
__license__ = "MIT"
__date__ = "17/07/2018"


import logging

from silx.utils.weakref import WeakList

_logger = logging.getLogger(__name__)


# Notifier ####################################################################

class Notifier(object):
    """Base class for object with notification mechanism."""

    def __init__(self):
        self._listeners = WeakList()

    def addListener(self, listener):
        """Register a listener.

        Adding an already registered listener has no effect.

        :param callable listener: The function or method to register.
        """
        if listener not in self._listeners:
            self._listeners.append(listener)
        else:
            _logger.warning('Ignoring addition of an already registered listener')

    def removeListener(self, listener):
        """Remove a previously registered listener.

        :param callable listener: The function or method to unregister.
        """
        try:
            self._listeners.remove(listener)
        except ValueError:
            _logger.warning('Trying to remove a listener that is not registered')

    def notify(self, *args, **kwargs):
        """Notify all registered listeners with the given parameters.

        Listeners are called directly in this method.
        Listeners are called in the order they were registered.
        """
        for listener in self._listeners:
            listener(self, *args, **kwargs)


def notifyProperty(attrName, copy=False, converter=None, doc=None):
    """Create a property that adds notification to an attribute.

    :param str attrName: The name of the attribute to wrap.
    :param bool copy: Whether to return a copy of the attribute
                      or not (the default).
    :param converter: Function converting input value to appropriate type
                      This function takes a single argument and return the
                      converted value.
                      It can be used to perform some asserts.
    :param str doc: The docstring of the property
    :return: A property with getter and setter
    """
    if copy:
        def getter(self):
            return getattr(self, attrName).copy()
    else:
        def getter(self):
            return getattr(self, attrName)

    if converter is None:
        def setter(self, value):
            if getattr(self, attrName) != value:
                setattr(self, attrName, value)
                self.notify()

    else:
        def setter(self, value):
            value = converter(value)
            if getattr(self, attrName) != value:
                setattr(self, attrName, value)
                self.notify()

    return property(getter, setter, doc=doc)


class HookList(list):
    """List with hooks before and after modification."""

    def __init__(self, iterable):
        super(HookList, self).__init__(iterable)

        self._listWasChangedHook('__init__', iterable)

    def _listWillChangeHook(self, methodName, *args, **kwargs):
        """To override. Called before modifying the list.

        This method is called with the name of the method called to
        modify the list and its parameters.
        """
        pass

    def _listWasChangedHook(self, methodName, *args, **kwargs):
        """To override. Called after modifying the list.

        This method is called with the name of the method called to
        modify the list and its parameters.
        """
        pass

    # Wrapping methods that modify the list

    def _wrapper(self, methodName, *args, **kwargs):
        """Generic wrapper of list methods calling the hooks."""
        self._listWillChangeHook(methodName, *args, **kwargs)
        result = getattr(super(HookList, self),
                         methodName)(*args, **kwargs)
        self._listWasChangedHook(methodName, *args, **kwargs)
        return result

    # Add methods

    def __iadd__(self, *args, **kwargs):
        return self._wrapper('__iadd__', *args, **kwargs)

    def __imul__(self, *args, **kwargs):
        return self._wrapper('__imul__', *args, **kwargs)

    def append(self, *args, **kwargs):
        return self._wrapper('append', *args, **kwargs)

    def extend(self, *args, **kwargs):
        return self._wrapper('extend', *args, **kwargs)

    def insert(self, *args, **kwargs):
        return self._wrapper('insert', *args, **kwargs)

    # Remove methods

    def __delitem__(self, *args, **kwargs):
        return self._wrapper('__delitem__', *args, **kwargs)

    def __delslice__(self, *args, **kwargs):
        return self._wrapper('__delslice__', *args, **kwargs)

    def remove(self, *args, **kwargs):
        return self._wrapper('remove', *args, **kwargs)

    def pop(self, *args, **kwargs):
        return self._wrapper('pop', *args, **kwargs)

    # Set methods

    def __setitem__(self, *args, **kwargs):
        return self._wrapper('__setitem__', *args, **kwargs)

    def __setslice__(self, *args, **kwargs):
        return self._wrapper('__setslice__', *args, **kwargs)

    # In place methods

    def sort(self, *args, **kwargs):
        return self._wrapper('sort', *args, **kwargs)

    def reverse(self, *args, **kwargs):
        return self._wrapper('reverse', *args, **kwargs)


class NotifierList(HookList, Notifier):
    """List of Notifiers with notification mechanism.

    This class registers itself as a listener of the list items.

    The default listener method forward notification from list items
    to the listeners of the list.
    """

    def __init__(self, iterable=()):
        Notifier.__init__(self)
        HookList.__init__(self, iterable)

    def _listWillChangeHook(self, methodName, *args, **kwargs):
        for item in self:
            item.removeListener(self._notified)

    def _listWasChangedHook(self, methodName, *args, **kwargs):
        for item in self:
            item.addListener(self._notified)
        self.notify()

    def _notified(self, source, *args, **kwargs):
        """Default listener forwarding list item changes to its listeners."""
        # Avoid infinite recursion if the list is listening itself
        if source is not self:
            self.notify(*args, **kwargs)



# /*##########################################################################
# Copyright (C) 2018 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ############################################################################*/
"""
This module provides implementations based on marching squares algorithms.

The main implementation is done by :class:`MarchingSquaresMergeImpl`. It was
designed to speed up the computation of iso surface using Cython and OpenMP.
It also provides features like support of mask, and cache of min/max per tiles
which is very efficient to find many iso contours from image gradient.

Utilitary functions are provided as facade for simple use.
:meth:`find_contours` to find iso contours from an image and using the same
main signature as `find_contours` from `skimage`, but supporting mask.
And :meth:`find_pixels` which returns a set of pixel coords containing the
points of the iso contours.
"""

__authors__ = ["V. Valls"]
__license__ = "MIT"
__date__ = "02/07/2018"


from ._mergeimpl import MarchingSquaresMergeImpl


def _factory(engine, image, mask):
    """Factory to create the marching square implementation from the engine
    name"""
    if engine == "merge":
        return MarchingSquaresMergeImpl(image, mask)
    elif engine == "skimage":
        from _skimage import MarchingSquaresSciKitImage
        return MarchingSquaresSciKitImage(image, mask)
    else:
        raise ValueError("Engine '%s' is not supported ('merge' or 'skimage' expected).")


def find_pixels(image, level, mask=None):
    """
    Find the pixels following the iso contours at the given `level`.

    These pixels are localized by the bound of the segment generated by the
    iso contour algorithm.

    The result is returned as a numpy array storing a list of coordinates y/x.

    .. code-block:: python

        # Example using a mask
        shape = 100, 100
        image = numpy.random.random(shape)
        mask = numpy.random.random(shape) < 0.01
        pixels = silx.image.marchingsquares.find_pixels(image, 0.5, mask=mask)

    :param numpy.ndarray image: Image to process
    :param float level: Level of the requested iso contours.
    :param numpy.ndarray mask: An optional mask (a non-zero value invalidate
        the pixels of the image)
    :returns: An array of coordinates in y/x
    :rtype: numpy.ndarray
    """
    assert(image is not None)
    if mask is not None:
        assert(image.shape == mask.shape)
    engine = "merge"
    impl = _factory(engine, image, mask)
    return impl.find_pixels(level)


def find_contours(image, level, mask=None):
    """
    Find the iso contours at the given `level`.

    The result is returned as a list of polygons.

    .. code-block:: python

        # Example using a mask
        shape = 100, 100
        image = numpy.random.random(shape)
        mask = numpy.random.random(shape) < 0.01
        polygons = silx.image.marchingsquares.find_contours(image, 0.5, mask=mask)

    :param numpy.ndarray image: Image to process
    :param float level: Level of the requested iso contours.
    :param numpy.ndarray mask: An optional mask (a non-zero value invalidate
        the pixels of the image)
    :returns: A list of array containing y-x coordinates of points
    :rtype: List[numpy.ndarray]
    """
    assert(image is not None)
    if mask is not None:
        assert(image.shape == mask.shape)
    engine = "merge"
    impl = _factory(engine, image, mask)
    return impl.find_contours(level)



#!/usr/bin/env python
# /*##########################################################################
#
# Copyright (c) 2017-2019 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""This module contains library wide configuration.
"""

__authors__ = ["V. Valls"]
__license__ = "MIT"
__date__ = "09/11/2018"


class Config(object):
    """
    Class containing shared global configuration for the silx library.

    .. versionadded:: 0.8
    """

    DEFAULT_PLOT_BACKEND = "matplotlib", "opengl"
    """Default plot backend.

    It will be used as default backend for all the next created PlotWidget.

    This attribute can be set with:

    - 'matplotlib' (default) or 'mpl'
    - 'opengl', 'gl'
    - 'none'
    - A :class:`silx.gui.plot.backend.BackendBase.BackendBase` class
    - A callable returning backend class or binding name

    If multiple backends are provided, the first available one is used.

    .. versionadded:: 0.8
    """

    DEFAULT_COLORMAP_NAME = 'gray'
    """Default LUT for the plot widgets.

    The available list of names are available in the module
    :module:`silx.gui.colors`.

    .. versionadded:: 0.8
    """

    DEFAULT_PLOT_IMAGE_Y_AXIS_ORIENTATION = 'upward'
    """Default Y-axis orientation for plot widget displaying images.

    This attribute can be set with:

    - 'upward' (default), which set the origin to the bottom with an upward
        orientation.
    - 'downward', which set the origin to the top with a backward orientation.

    It will have an influence on:

    - :class:`silx.gui.plot.StackWidget`
    - :class:`silx.gui.plot.ComplexImageView`
    - :class:`silx.gui.plot.Plot2D`
    - :class:`silx.gui.plot.ImageView`

    .. versionadded:: 0.8
    """

    DEFAULT_PLOT_CURVE_COLORS = ['#000000',  # black
                                 '#0000ff',  # blue
                                 '#ff0000',  # red
                                 '#00ff00',  # green
                                 '#ff66ff',  # pink
                                 '#ffff00',  # yellow
                                 '#a52a2a',  # brown
                                 '#00ffff',  # cyan
                                 '#ff00ff',  # magenta
                                 '#ff9900',  # orange
                                 '#6600ff',  # violet
                                 '#a0a0a4',  # grey
                                 '#000080',  # darkBlue
                                 '#800000',  # darkRed
                                 '#008000',  # darkGreen
                                 '#008080',  # darkCyan
                                 '#800080',  # darkMagenta
                                 '#808000',  # darkYellow
                                 '#660000']  # darkBrown
    """Default list of colors for plot widget displaying curves.

    It will have an influence on:

    - :class:`silx.gui.plot.PlotWidget`

    .. versionadded:: 0.9
    """

    DEFAULT_PLOT_CURVE_SYMBOL_MODE = False
    """Whether to display curves with markers or not by default in PlotWidget.

    It will have an influence on PlotWidget curve items.

    .. versionadded:: 0.10
    """

    DEFAULT_PLOT_SYMBOL = 'o'
    """Default marker of the item.

    It will have an influence on PlotWidget items

    Supported symbols:
    
        - 'o', 'Circle'
        - 'd', 'Diamond'
        - 's', 'Square'
        - '+', 'Plus'
        - 'x', 'Cross'
        - '.', 'Point'
        - ',', 'Pixel'
        - '',  'None'

    .. versionadded:: 0.10
    """

    DEFAULT_PLOT_SYMBOL_SIZE = 6.0
    """Default marker size of the item.

    It will have an influence on PlotWidget items

    .. versionadded:: 0.10
    """



# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""Tests for CompareImages widget"""

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "23/07/2018"

import unittest
import numpy
import weakref

from silx.gui.utils.testutils import TestCaseQt
from silx.gui.plot.CompareImages import CompareImages


class TestCompareImages(TestCaseQt):
    """Test that CompareImages widget is working in some cases"""

    def setUp(self):
        super(TestCompareImages, self).setUp()
        self.widget = CompareImages()

    def tearDown(self):
        ref = weakref.ref(self.widget)
        self.widget = None
        self.qWaitForDestroy(ref)
        super(TestCompareImages, self).tearDown()

    def testIntensityImage(self):
        image1 = numpy.random.rand(10, 10)
        image2 = numpy.random.rand(10, 10)
        self.widget.setData(image1, image2)

    def testRgbImage(self):
        image1 = numpy.random.randint(0, 255, size=(10, 10, 3))
        image2 = numpy.random.randint(0, 255, size=(10, 10, 3))
        self.widget.setData(image1, image2)

    def testRgbaImage(self):
        image1 = numpy.random.randint(0, 255, size=(10, 10, 4))
        image2 = numpy.random.randint(0, 255, size=(10, 10, 4))
        self.widget.setData(image1, image2)

    def testVizualisations(self):
        image1 = numpy.random.rand(10, 10)
        image2 = numpy.random.rand(10, 10)
        self.widget.setData(image1, image2)
        for mode in CompareImages.VisualizationMode:
            self.widget.setVisualizationMode(mode)

    def testAlignemnt(self):
        image1 = numpy.random.rand(10, 10)
        image2 = numpy.random.rand(5, 5)
        self.widget.setData(image1, image2)
        for mode in CompareImages.AlignmentMode:
            self.widget.setAlignmentMode(mode)

    def testGetPixel(self):
        image1 = numpy.random.rand(11, 11)
        image2 = numpy.random.rand(5, 5)
        image1[5, 5] = 111.111
        image2[2, 2] = 222.222
        self.widget.setData(image1, image2)
        expectedValue = {}
        expectedValue[CompareImages.AlignmentMode.CENTER] = 222.222
        expectedValue[CompareImages.AlignmentMode.STRETCH] = 222.222
        expectedValue[CompareImages.AlignmentMode.ORIGIN] = None
        for mode in expectedValue.keys():
            self.widget.setAlignmentMode(mode)
            data = self.widget.getRawPixelData(11 / 2.0, 11 / 2.0)
            data1, data2 = data
            self.assertEqual(data1, 111.111)
            self.assertEqual(data2, expectedValue[mode])

    def testImageEmpty(self):
        self.widget.setData(image1=None, image2=None)
        self.assertTrue(self.widget.getRawPixelData(11 / 2.0, 11 / 2.0) == (None, None))

    def testSetImageSeparately(self):
        self.widget.setImage1(numpy.random.rand(10, 10))
        self.widget.setImage2(numpy.random.rand(10, 10))
        for mode in CompareImages.VisualizationMode:
            self.widget.setVisualizationMode(mode)



# /*##########################################################################
#
# Copyright (c) 2020 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""Tests for the silx.gui.utils.glutils module."""

__authors__ = ["T. Vincent"]
__license__ = "MIT"
__date__ = "15/01/2020"


import logging
import pytest

from silx.gui.utils.glutils import isOpenGLAvailable


_logger = logging.getLogger(__name__)


@pytest.mark.parametrize("params", (((2, 1), False), ((2, 1), False), ((1000, 1), False), ((2, 1), True)))
def testOpenGLAvailable(params):
    version, shareOpenGLContexts = params
    result = isOpenGLAvailable(version=version, shareOpenGLContexts=shareOpenGLContexts)
    _logger.info("isOpenGLAvailable returned: %s", str(result))
    if version[0] == 1000:
        assert not result
    if not result:
        assert not result.status
        assert len(result.error) > 0
    else:
        assert result.status
        assert len(result.error) == 0



# /*##########################################################################
#
# Copyright (c) 2016-2020 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""Basic tests for PlotWindow"""

__authors__ = ["T. Vincent"]
__license__ = "MIT"
__date__ = "27/06/2017"


import unittest
import numpy
import pytest

from silx.gui.utils.testutils import TestCaseQt, getQToolButtonFromAction

from silx.gui import qt
from silx.gui.plot import PlotWindow
from silx.gui.colors import Colormap


class TestPlotWindow(TestCaseQt):
    """Base class for tests of PlotWindow."""

    def setUp(self):
        super(TestPlotWindow, self).setUp()
        self.plot = PlotWindow()
        self.plot.show()
        self.qWaitForWindowExposed(self.plot)

    def tearDown(self):
        self.plot.setAttribute(qt.Qt.WA_DeleteOnClose)
        self.plot.close()
        del self.plot
        super(TestPlotWindow, self).tearDown()

    def testActions(self):
        """Test the actions QToolButtons"""
        self.plot.setLimits(1, 100, 1, 100)

        checkList = [  # QAction, Plot state getter
            (self.plot.xAxisAutoScaleAction, self.plot.getXAxis().isAutoScale),
            (self.plot.yAxisAutoScaleAction, self.plot.getYAxis().isAutoScale),
            (self.plot.xAxisLogarithmicAction, self.plot.getXAxis()._isLogarithmic),
            (self.plot.yAxisLogarithmicAction, self.plot.getYAxis()._isLogarithmic),
            (self.plot.gridAction, self.plot.getGraphGrid),
        ]

        for action, getter in checkList:
            self.mouseMove(self.plot)
            initialState = getter()
            toolButton = getQToolButtonFromAction(action)
            self.assertIsNot(toolButton, None)
            self.mouseClick(toolButton, qt.Qt.LeftButton)
            self.assertNotEqual(getter(), initialState,
                                msg='"%s" state not changed' % action.text())

            self.mouseClick(toolButton, qt.Qt.LeftButton)
            self.assertEqual(getter(), initialState,
                             msg='"%s" state not changed' % action.text())

        # Trigger a zoom reset
        self.mouseMove(self.plot)
        resetZoomAction = self.plot.resetZoomAction
        toolButton = getQToolButtonFromAction(resetZoomAction)
        self.assertIsNot(toolButton, None)
        self.mouseClick(toolButton, qt.Qt.LeftButton)

    def testDockWidgets(self):
        """Test add/remove dock widgets"""
        dock1 = qt.QDockWidget('Test 1')
        dock1.setWidget(qt.QLabel('Test 1'))

        self.plot.addTabbedDockWidget(dock1)
        self.qapp.processEvents()

        self.plot.removeDockWidget(dock1)
        self.qapp.processEvents()

        dock2 = qt.QDockWidget('Test 2')
        dock2.setWidget(qt.QLabel('Test 2'))

        self.plot.addTabbedDockWidget(dock2)
        self.qapp.processEvents()

        if qt.BINDING != 'PySide2':
            # Weird bug with PySide2 later upon gc.collect() when getting the layout
            self.assertNotEqual(self.plot.layout().indexOf(dock2),
                                -1,
                                "dock2 not properly displayed")

    def testToolAspectRatio(self):
        self.plot.toolBar()
        self.plot.keepDataAspectRatioButton.keepDataAspectRatio()
        self.assertTrue(self.plot.isKeepDataAspectRatio())
        self.plot.keepDataAspectRatioButton.dontKeepDataAspectRatio()
        self.assertFalse(self.plot.isKeepDataAspectRatio())

    def testToolYAxisOrigin(self):
        self.plot.toolBar()
        self.plot.yAxisInvertedButton.setYAxisUpward()
        self.assertFalse(self.plot.getYAxis().isInverted())
        self.plot.yAxisInvertedButton.setYAxisDownward()
        self.assertTrue(self.plot.getYAxis().isInverted())

    def testColormapAutoscaleCache(self):
        # Test that the min/max cache is not computed twice

        old = Colormap._computeAutoscaleRange
        self._count = 0
        def _computeAutoscaleRange(colormap, data):
            self._count = self._count + 1
            return 10, 20
        Colormap._computeAutoscaleRange = _computeAutoscaleRange
        try:
            colormap = Colormap(name='red')
            self.plot.setVisible(True)

            # Add an image
            data = numpy.arange(8**2).reshape(8, 8)
            self.plot.addImage(data, legend="foo", colormap=colormap)
            self.plot.setActiveImage("foo")

            # Use the colorbar
            self.plot.getColorBarWidget().setVisible(True)
            self.qWait(50)

            # Remove and add again the same item
            image = self.plot.getImage("foo")
            self.plot.removeImage("foo")
            self.plot.addItem(image)
            self.qWait(50)
        finally:
            Colormap._computeAutoscaleRange = old
        self.assertEqual(self._count, 1)
        del self._count

    @pytest.mark.usefixtures("use_opengl")
    def testSwitchBackend(self):
        """Test switching an empty plot"""
        self.plot.resetZoom()
        xlimits = self.plot.getXAxis().getLimits()
        ylimits = self.plot.getYAxis().getLimits()
        isKeepAspectRatio = self.plot.isKeepDataAspectRatio()

        for backend in ('gl', 'mpl'):
            with self.subTest():
                self.plot.setBackend(backend)
                self.plot.replot()
                self.assertEqual(self.plot.getXAxis().getLimits(), xlimits)
                self.assertEqual(self.plot.getYAxis().getLimits(), ylimits)
                self.assertEqual(
                    self.plot.isKeepDataAspectRatio(), isKeepAspectRatio)



#!/usr/bin/env python
# /*##########################################################################
#
# Copyright (c) 2016-2021 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""
Example for the use of the ImageFileDialog.
"""

__authors__ = ["V. Valls"]
__license__ = "MIT"
__date__ = "14/02/2018"

import enum
import logging
from silx.gui import qt
from silx.gui.dialog.ImageFileDialog import ImageFileDialog
from silx.gui.dialog.DataFileDialog import DataFileDialog
import silx.io


logging.basicConfig()


class Mode(enum.Enum):
    DEFAULT_FILEDIALOG = 0
    IMAGEFILEDIALOG = 1
    DATAFILEDIALOG = 2
    DATAFILEDIALOG_DATASET = 3
    DATAFILEDIALOG_GROUP = 4
    DATAFILEDIALOG_NXENTRY = 5


class DialogExample(qt.QMainWindow):

    def __init__(self, parent=None):
        super(DialogExample, self).__init__(parent)

        self.__state = {}

        centralWidget = qt.QWidget(self)
        layout = qt.QHBoxLayout()
        centralWidget.setLayout(layout)

        options = self.createOptions()
        layout.addWidget(options)

        buttonGroup = qt.QGroupBox()
        buttonGroup.setTitle("Create dialog")
        layout.addWidget(buttonGroup)
        buttonLayout = qt.QVBoxLayout()
        buttonGroup.setLayout(buttonLayout)

        # ImageFileDialog

        b1 = qt.QPushButton(self)
        b1.setMinimumHeight(50)
        b1.setText("Open a dialog")
        b1.clicked.connect(self.openDialog)
        buttonLayout.addWidget(b1)

        b2 = qt.QPushButton(self)
        b2.setMinimumHeight(50)
        b2.setText("Open a dialog with state stored")
        b2.clicked.connect(self.openDialogStoredState)
        buttonLayout.addWidget(b2)

        b3 = qt.QPushButton(self)
        b3.setMinimumHeight(50)
        b3.setText("Open a dialog at home")
        b3.clicked.connect(self.openDialogAtHome)
        buttonLayout.addWidget(b3)

        b4 = qt.QPushButton(self)
        b4.setMinimumHeight(50)
        b4.setText("Open a dialog at computer root")
        b4.clicked.connect(self.openDialogAtComputer)
        buttonLayout.addWidget(b4)

        self.setCentralWidget(centralWidget)

    def createOptions(self):
        panel = qt.QGroupBox()
        panel.setTitle("Options")
        layout = qt.QVBoxLayout()
        panel.setLayout(layout)
        group = qt.QButtonGroup(panel)

        radio = qt.QRadioButton(panel)
        radio.setText("Qt QFileDialog")
        radio.setProperty("Mode", Mode.DEFAULT_FILEDIALOG)
        group.addButton(radio)
        layout.addWidget(radio)

        radio = qt.QRadioButton(panel)
        radio.setText("silx ImageFileDialog")
        radio.setProperty("Mode", Mode.IMAGEFILEDIALOG)
        group.addButton(radio)
        layout.addWidget(radio)

        radio = qt.QRadioButton(panel)
        radio.setChecked(True)
        radio.setText("silx DataFileDialog")
        radio.setProperty("Mode", Mode.DATAFILEDIALOG)
        group.addButton(radio)
        layout.addWidget(radio)

        radio = qt.QRadioButton(panel)
        radio.setText("silx DataFileDialog (filter=dataset)")
        radio.setProperty("Mode", Mode.DATAFILEDIALOG_DATASET)
        group.addButton(radio)
        layout.addWidget(radio)

        radio = qt.QRadioButton(panel)
        radio.setText("silx DataFileDialog (filter=group)")
        radio.setProperty("Mode", Mode.DATAFILEDIALOG_GROUP)
        group.addButton(radio)
        layout.addWidget(radio)

        radio = qt.QRadioButton(panel)
        radio.setText("silx DataFileDialog (filter=NXentry)")
        radio.setProperty("Mode", Mode.DATAFILEDIALOG_NXENTRY)
        group.addButton(radio)
        layout.addWidget(radio)

        self.__options = group
        return panel

    def printResult(self, dialog, result):
        if not result:
            print("Nothing selected")
            return

        print("Selection:")
        if isinstance(dialog, qt.QFileDialog):
            print("- Files: %s" % dialog.selectedFiles())
        elif isinstance(dialog, ImageFileDialog):
            print("- File: %s" % dialog.selectedFile())
            print("- URL: %s" % dialog.selectedUrl())
            print("- Data URL: %s" % dialog.selectedDataUrl())
            image = dialog.selectedImage()
            print("- Image: <dtype: %s, shape: %s>" % (image.dtype, image.shape))
        elif isinstance(dialog, DataFileDialog):
            print("- File: %s" % dialog.selectedFile())
            print("- URL: %s" % dialog.selectedUrl())
            print("- Data URL: %s" % dialog.selectedDataUrl())
            try:
                data = dialog.selectedData()
                print("- Data: <dtype: %s, shape: %s>" % (data.dtype, data.shape))
            except Exception as e:
                print("- Data: %s" % e)

            url = dialog.selectedDataUrl()
            with silx.io.open(url.file_path()) as h5:
                node = h5[url.data_path()]
                print("- Node: %s" % node)
        else:
            assert(False)

    def createDialog(self):
        print("")
        print("-------------------------")
        print("----- Create dialog -----")
        print("-------------------------")
        button = self.__options.checkedButton()
        mode = button.property("Mode")
        if mode == Mode.DEFAULT_FILEDIALOG:
            dialog = qt.QFileDialog(self)
            dialog.setAcceptMode(qt.QFileDialog.AcceptOpen)
        elif mode == Mode.IMAGEFILEDIALOG:
            dialog = ImageFileDialog(self)
        elif mode == Mode.DATAFILEDIALOG:
            dialog = DataFileDialog(self)
        elif mode == Mode.DATAFILEDIALOG_DATASET:
            dialog = DataFileDialog(self)
            dialog.setFilterMode(DataFileDialog.FilterMode.ExistingDataset)
        elif mode == Mode.DATAFILEDIALOG_GROUP:
            dialog = DataFileDialog(self)
            dialog.setFilterMode(DataFileDialog.FilterMode.ExistingGroup)
        elif mode == Mode.DATAFILEDIALOG_NXENTRY:
            def customFilter(obj):
                if "NX_class" in obj.attrs:
                    return obj.attrs["NX_class"] in [b"NXentry", u"NXentry"]
                return False
            dialog = DataFileDialog(self)
            dialog.setFilterMode(DataFileDialog.FilterMode.ExistingGroup)
            dialog.setFilterCallback(customFilter)
        else:
            assert(False)
        return dialog

    def openDialog(self):
        # Clear the dialog
        dialog = self.createDialog()

        # Execute the dialog as modal
        result = dialog.exec()
        self.printResult(dialog, result)

    def openDialogStoredState(self):
        # Clear the dialog
        dialog = self.createDialog()
        if dialog.__class__ in self.__state:
            dialog.restoreState(self.__state[dialog.__class__])

        # Execute the dialog as modal
        result = dialog.exec()
        self.__state[dialog.__class__] = dialog.saveState()
        self.printResult(dialog, result)

    def openDialogAtHome(self):
        # Clear the dialog
        path = qt.QDir.homePath()
        dialog = self.createDialog()
        dialog.setDirectory(path)

        # Execute the dialog as modal
        result = dialog.exec()
        self.printResult(dialog, result)

    def openDialogAtComputer(self):
        # Clear the dialog
        path = ""
        dialog = self.createDialog()
        dialog.setDirectory(path)

        # Execute the dialog as modal
        result = dialog.exec()
        self.printResult(dialog, result)


def main():
    app = qt.QApplication([])
    example = DialogExample()
    example.show()
    app.exec()


if __name__ == "__main__":
    main()



# /*##########################################################################
# Copyright (C) 2018 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ############################################################################*/

__authors__ = ["V. Valls"]
__license__ = "MIT"
__date__ = "05/04/2018"


import numpy
import skimage.measure


class MarchingSquaresSciKitImage(object):
    """Reference implementation of a marching squares using sci-kit image.

    It uses `skimage.measure.find_contours` to find iso contours taking care of
    an optional mask. As result the computation is not accurate but can be used
    as reference for benchmark or for testing the API without compiling the
    cython part of silx.

    :param numpy.ndarray image: 2d-image containing the values
    :param numpy.ndarray mask: Optional 2d-image containing mask to cancel
        mask on part of the image. A `0` means the pixel at this location is
        valid, else the pixel from the image will not be used.
    """

    def __init__(self, image, mask=None):
        self._image = image
        self._mask = mask

    _deltas = [(0.0, 0.0), (0.99, 0.0), (0.0, 0.99), (0.99, 0.99)]

    def _flag_coord_over_mask(self, coord):
        """Flag coord over the mask as NaN"""
        for dx, dy in self._deltas:
            if self._mask[int(coord[0] + dx), int(coord[1] + dy)] != 0:
                return float("nan"), float("nan")
        return coord

    def find_pixels(self, level):
        """
        Compute the pixels from the image over the requested iso contours
        at this `level`.

        This implementation have to use `skimage.measure.find_contours` then
        it is not accurate nor efficient.

        :param float level: Level of the requested iso contours.
        :returns: An array of y-x coordinates.
        :rtype: numpy.ndarray
        """
        polylines = skimage.measure.find_contours(self._image, level=level)
        size = 0
        for polyline in polylines:
            size += len(polyline)
        result = numpy.empty((size, 2), dtype=numpy.int32)
        size = 0
        delta = numpy.array([0.5, 0.5])
        for polyline in polylines:
            if len(polyline) == 0:
                continue
            integer_polyline = numpy.floor(polyline + delta)
            result[size:size + len(polyline)] = integer_polyline
            size += len(polyline)

        if len(result) == 0:
            return result

        if self._mask is not None:
            # filter out pixels over the mask
            x_dim = self._image.shape[1]
            indexes = result[:, 0] * x_dim + result[:, 1]
            indexes = indexes.ravel()
            mask = self._mask.ravel()
            indexes = numpy.unique(indexes)
            indexes = indexes[mask[indexes] == 0]
            pixels = numpy.concatenate((indexes // x_dim, indexes % x_dim))
            pixels.shape = 2, -1
            pixels = pixels.T
            result = pixels
        else:
            # Note: Cound be done using a single line numpy.unique(result, axis=0)
            # But here it supports Debian 8
            x_dim = self._image.shape[1]
            indexes = result[:, 0] * x_dim + result[:, 1]
            indexes = indexes.ravel()
            indexes = numpy.unique(indexes)
            pixels = numpy.concatenate((indexes // x_dim, indexes % x_dim))
            pixels.shape = 2, -1
            pixels = pixels.T
            result = pixels
        return result

    def find_contours(self, level):
        """
        Compute the list of polygons of the iso contours at this `level`.

        If no mask is involved, the result is the same as
        `skimage.measure.find_contours`.

        If the result have to be filtered with a mask, the result is not
        accurate nor efficient. Polygons are not splited, but only points are
        filtered out using NaN coordinates. This could create artefacts.

        :param float level: Level of the requested iso contours.
        :returns: A list of array containg y-x coordinates of points
        :rtype: List[numpy.ndarray]
        """
        polylines = skimage.measure.find_contours(self._image, level=level)
        if self._mask is None:
            return polylines
        result = []
        for polyline in polylines:
            polyline = map(self._flag_coord_over_mask, polyline)
            polyline = list(polyline)
            polyline = numpy.array(polyline)
            result.append(polyline)
        return result



# /*##########################################################################
#
# Copyright (c) 2004-2020 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""This module provides a set of QToolButton to use with
:class:`~silx.gui.plot.PlotWidget`.

The following QToolButton are available:

- :class:`.AspectToolButton`
- :class:`.YAxisOriginToolButton`
- :class:`.ProfileToolButton`
- :class:`.SymbolToolButton`

"""

__authors__ = ["V. Valls", "H. Payno"]
__license__ = "MIT"
__date__ = "27/06/2017"


import functools
import logging
import weakref

from .. import icons
from .. import qt
from ... import config

from .items import SymbolMixIn, Scatter


_logger = logging.getLogger(__name__)


class PlotToolButton(qt.QToolButton):
    """A QToolButton connected to a :class:`~silx.gui.plot.PlotWidget`.
    """

    def __init__(self, parent=None, plot=None):
        super(PlotToolButton, self).__init__(parent)
        self._plotRef = None
        if plot is not None:
            self.setPlot(plot)

    def plot(self):
        """
        Returns the plot connected to the widget.
        """
        return None if self._plotRef is None else self._plotRef()

    def setPlot(self, plot):
        """
        Set the plot connected to the widget

        :param plot: :class:`.PlotWidget` instance on which to operate.
        """
        previousPlot = self.plot()

        if previousPlot is plot:
            return
        if previousPlot is not None:
            self._disconnectPlot(previousPlot)

        if plot is None:
            self._plotRef = None
        else:
            self._plotRef = weakref.ref(plot)
            self._connectPlot(plot)

    def _connectPlot(self, plot):
        """
        Called when the plot is connected to the widget

        :param plot: :class:`.PlotWidget` instance
        """
        pass

    def _disconnectPlot(self, plot):
        """
        Called when the plot is disconnected from the widget

        :param plot: :class:`.PlotWidget` instance
        """
        pass


class AspectToolButton(PlotToolButton):
    """Tool button to switch keep aspect ratio of a plot"""

    STATE = None
    """Lazy loaded states used to feed AspectToolButton"""

    def __init__(self, parent=None, plot=None):
        if self.STATE is None:
            self.STATE = {}
            # dont keep ratio
            self.STATE[False, "icon"] = icons.getQIcon('shape-ellipse-solid')
            self.STATE[False, "state"] = "Aspect ratio is not kept"
            self.STATE[False, "action"] = "Do no keep data aspect ratio"
            # keep ratio
            self.STATE[True, "icon"] = icons.getQIcon('shape-circle-solid')
            self.STATE[True, "state"] = "Aspect ratio is kept"
            self.STATE[True, "action"] = "Keep data aspect ratio"

        super(AspectToolButton, self).__init__(parent=parent, plot=plot)

        keepAction = self._createAction(True)
        keepAction.triggered.connect(self.keepDataAspectRatio)
        keepAction.setIconVisibleInMenu(True)

        dontKeepAction = self._createAction(False)
        dontKeepAction.triggered.connect(self.dontKeepDataAspectRatio)
        dontKeepAction.setIconVisibleInMenu(True)

        menu = qt.QMenu(self)
        menu.addAction(keepAction)
        menu.addAction(dontKeepAction)
        self.setMenu(menu)
        self.setPopupMode(qt.QToolButton.InstantPopup)

    def _createAction(self, keepAspectRatio):
        icon = self.STATE[keepAspectRatio, "icon"]
        text = self.STATE[keepAspectRatio, "action"]
        return qt.QAction(icon, text, self)

    def _connectPlot(self, plot):
        plot.sigSetKeepDataAspectRatio.connect(self._keepDataAspectRatioChanged)
        self._keepDataAspectRatioChanged(plot.isKeepDataAspectRatio())

    def _disconnectPlot(self, plot):
        plot.sigSetKeepDataAspectRatio.disconnect(self._keepDataAspectRatioChanged)

    def keepDataAspectRatio(self):
        """Configure the plot to keep the aspect ratio"""
        plot = self.plot()
        if plot is not None:
            # This will trigger _keepDataAspectRatioChanged
            plot.setKeepDataAspectRatio(True)

    def dontKeepDataAspectRatio(self):
        """Configure the plot to not keep the aspect ratio"""
        plot = self.plot()
        if plot is not None:
            # This will trigger _keepDataAspectRatioChanged
            plot.setKeepDataAspectRatio(False)

    def _keepDataAspectRatioChanged(self, aspectRatio):
        """Handle Plot set keep aspect ratio signal"""
        icon, toolTip = self.STATE[aspectRatio, "icon"], self.STATE[aspectRatio, "state"]
        self.setIcon(icon)
        self.setToolTip(toolTip)


class YAxisOriginToolButton(PlotToolButton):
    """Tool button to switch the Y axis orientation of a plot."""

    STATE = None
    """Lazy loaded states used to feed YAxisOriginToolButton"""

    def __init__(self, parent=None, plot=None):
        if self.STATE is None:
            self.STATE = {}
            # is down
            self.STATE[False, "icon"] = icons.getQIcon('plot-ydown')
            self.STATE[False, "state"] = "Y-axis is oriented downward"
            self.STATE[False, "action"] = "Orient Y-axis downward"
            # keep ration
            self.STATE[True, "icon"] = icons.getQIcon('plot-yup')
            self.STATE[True, "state"] = "Y-axis is oriented upward"
            self.STATE[True, "action"] = "Orient Y-axis upward"

        super(YAxisOriginToolButton, self).__init__(parent=parent, plot=plot)

        upwardAction = self._createAction(True)
        upwardAction.triggered.connect(self.setYAxisUpward)
        upwardAction.setIconVisibleInMenu(True)

        downwardAction = self._createAction(False)
        downwardAction.triggered.connect(self.setYAxisDownward)
        downwardAction.setIconVisibleInMenu(True)

        menu = qt.QMenu(self)
        menu.addAction(upwardAction)
        menu.addAction(downwardAction)
        self.setMenu(menu)
        self.setPopupMode(qt.QToolButton.InstantPopup)

    def _createAction(self, isUpward):
        icon = self.STATE[isUpward, "icon"]
        text = self.STATE[isUpward, "action"]
        return qt.QAction(icon, text, self)

    def _connectPlot(self, plot):
        yAxis = plot.getYAxis()
        yAxis.sigInvertedChanged.connect(self._yAxisInvertedChanged)
        self._yAxisInvertedChanged(yAxis.isInverted())

    def _disconnectPlot(self, plot):
        plot.getYAxis().sigInvertedChanged.disconnect(self._yAxisInvertedChanged)

    def setYAxisUpward(self):
        """Configure the plot to use y-axis upward"""
        plot = self.plot()
        if plot is not None:
            # This will trigger _yAxisInvertedChanged
            plot.getYAxis().setInverted(False)

    def setYAxisDownward(self):
        """Configure the plot to use y-axis downward"""
        plot = self.plot()
        if plot is not None:
            # This will trigger _yAxisInvertedChanged
            plot.getYAxis().setInverted(True)

    def _yAxisInvertedChanged(self, inverted):
        """Handle Plot set y axis inverted signal"""
        isUpward = not inverted
        icon, toolTip = self.STATE[isUpward, "icon"], self.STATE[isUpward, "state"]
        self.setIcon(icon)
        self.setToolTip(toolTip)


class ProfileOptionToolButton(PlotToolButton):
    """Button to define option on the profile"""
    sigMethodChanged = qt.Signal(str)
    
    def __init__(self, parent=None, plot=None):
        PlotToolButton.__init__(self, parent=parent, plot=plot)

        self.STATE = {}
        # is down
        self.STATE['sum', "icon"] = icons.getQIcon('math-sigma')
        self.STATE['sum', "state"] = "Compute profile sum"
        self.STATE['sum', "action"] = "Compute profile sum"
        # keep ration
        self.STATE['mean', "icon"] = icons.getQIcon('math-mean')
        self.STATE['mean', "state"] = "Compute profile mean"
        self.STATE['mean', "action"] = "Compute profile mean"

        self.sumAction = self._createAction('sum')
        self.sumAction.triggered.connect(self.setSum)
        self.sumAction.setIconVisibleInMenu(True)
        self.sumAction.setCheckable(True)
        self.sumAction.setChecked(True)

        self.meanAction = self._createAction('mean')
        self.meanAction.triggered.connect(self.setMean)
        self.meanAction.setIconVisibleInMenu(True)
        self.meanAction.setCheckable(True)

        menu = qt.QMenu(self)
        menu.addAction(self.sumAction)
        menu.addAction(self.meanAction)
        self.setMenu(menu)
        self.setPopupMode(qt.QToolButton.InstantPopup)
        self._method = 'mean'
        self._update()

    def _createAction(self, method):
        icon = self.STATE[method, "icon"]
        text = self.STATE[method, "action"]
        return qt.QAction(icon, text, self)

    def setSum(self):
        self.setMethod('sum')

    def _update(self):
        icon = self.STATE[self._method, "icon"]
        toolTip = self.STATE[self._method, "state"]
        self.setIcon(icon)
        self.setToolTip(toolTip)
        self.sumAction.setChecked(self._method == "sum")
        self.meanAction.setChecked(self._method == "mean")

    def setMean(self):
        self.setMethod('mean')

    def setMethod(self, method):
        """Set the method to use.

        :param str method: Either 'sum' or 'mean'
        """
        if method != self._method:
            if method in ('sum', 'mean'):
                self._method = method
                self.sigMethodChanged.emit(self._method)
                self._update()
            else:
                _logger.warning(
                    "Unsupported method '%s'. Setting ignored.", method)

    def getMethod(self):
        """Returns the current method in use (See :meth:`setMethod`).

        :rtype: str
        """
        return self._method


class ProfileToolButton(PlotToolButton):
    """Button used in Profile3DToolbar to switch between 2D profile
    and 1D profile."""
    STATE = None
    """Lazy loaded states used to feed ProfileToolButton"""

    sigDimensionChanged = qt.Signal(int)

    def __init__(self, parent=None, plot=None):
        if self.STATE is None:
            self.STATE = {
                (1, "icon"): icons.getQIcon('profile1D'),
                (1, "state"): "1D profile is computed on visible image",
                (1, "action"): "1D profile on visible image",
                (2, "icon"): icons.getQIcon('profile2D'),
                (2, "state"): "2D profile is computed, one 1D profile for each image in the stack",
                (2, "action"): "2D profile on image stack"}
            # Compute 1D profile
            # Compute 2D profile

        super(ProfileToolButton, self).__init__(parent=parent, plot=plot)

        self._dimension = 1

        profile1DAction = self._createAction(1)
        profile1DAction.triggered.connect(self.computeProfileIn1D)
        profile1DAction.setIconVisibleInMenu(True)
        profile1DAction.setCheckable(True)
        profile1DAction.setChecked(True)
        self._profile1DAction = profile1DAction

        profile2DAction = self._createAction(2)
        profile2DAction.triggered.connect(self.computeProfileIn2D)
        profile2DAction.setIconVisibleInMenu(True)
        profile2DAction.setCheckable(True)
        self._profile2DAction = profile2DAction

        menu = qt.QMenu(self)
        menu.addAction(profile1DAction)
        menu.addAction(profile2DAction)
        self.setMenu(menu)
        self.setPopupMode(qt.QToolButton.InstantPopup)
        menu.setTitle('Select profile dimension')
        self.computeProfileIn1D()

    def _createAction(self, profileDimension):
        icon = self.STATE[profileDimension, "icon"]
        text = self.STATE[profileDimension, "action"]
        return qt.QAction(icon, text, self)

    def _profileDimensionChanged(self, profileDimension):
        """Update icon in toolbar, emit number of dimensions for profile"""
        self.setIcon(self.STATE[profileDimension, "icon"])
        self.setToolTip(self.STATE[profileDimension, "state"])
        self._dimension = profileDimension
        self.sigDimensionChanged.emit(profileDimension)
        self._profile1DAction.setChecked(profileDimension == 1)
        self._profile2DAction.setChecked(profileDimension == 2)

    def computeProfileIn1D(self):
        self._profileDimensionChanged(1)

    def computeProfileIn2D(self):
        self._profileDimensionChanged(2)

    def setDimension(self, dimension):
        """Set the selected dimension"""
        assert dimension in [1, 2]
        if self._dimension == dimension:
            return
        if dimension == 1:
            self.computeProfileIn1D()
        elif dimension == 2:
            self.computeProfileIn2D()
        else:
            _logger.warning("Unsupported dimension '%s'. Setting ignored.", dimension)

    def getDimension(self):
        """Get the selected dimension.

        :rtype: int (1 or 2)
        """
        return self._dimension


class _SymbolToolButtonBase(PlotToolButton):
    """Base class for PlotToolButton setting marker and size.

    :param parent: See QWidget
    :param plot: The `~silx.gui.plot.PlotWidget` to control
    """

    def __init__(self, parent=None, plot=None):
        super(_SymbolToolButtonBase, self).__init__(parent=parent, plot=plot)

    def _addSizeSliderToMenu(self, menu):
        """Add a slider to set size to the given menu

        :param QMenu menu:
        """
        slider = qt.QSlider(qt.Qt.Horizontal)
        slider.setRange(1, 20)
        slider.setValue(int(config.DEFAULT_PLOT_SYMBOL_SIZE))
        slider.setTracking(False)
        slider.valueChanged.connect(self._sizeChanged)
        widgetAction = qt.QWidgetAction(menu)
        widgetAction.setDefaultWidget(slider)
        menu.addAction(widgetAction)

    def _addSymbolsToMenu(self, menu):
        """Add symbols to the given menu

        :param QMenu menu:
        """
        for marker, name in zip(SymbolMixIn.getSupportedSymbols(),
                                SymbolMixIn.getSupportedSymbolNames()):
            action = qt.QAction(name, menu)
            action.setCheckable(False)
            action.triggered.connect(
                functools.partial(self._markerChanged, marker))
            menu.addAction(action)

    def _sizeChanged(self, value):
        """Manage slider value changed

        :param int value: Marker size
        """
        plot = self.plot()
        if plot is None:
            return

        for item in plot.getItems():
            if isinstance(item, SymbolMixIn):
                item.setSymbolSize(value)

    def _markerChanged(self, marker):
        """Manage change of marker.

        :param str marker: Letter describing the marker
        """
        plot = self.plot()
        if plot is None:
            return

        for item in plot.getItems():
            if isinstance(item, SymbolMixIn):
                item.setSymbol(marker)


class SymbolToolButton(_SymbolToolButtonBase):
    """A tool button with a drop-down menu to control symbol size and marker.

    :param parent: See QWidget
    :param plot: The `~silx.gui.plot.PlotWidget` to control
    """

    def __init__(self, parent=None, plot=None):
        super(SymbolToolButton, self).__init__(parent=parent, plot=plot)

        self.setToolTip('Set symbol size and marker')
        self.setIcon(icons.getQIcon('plot-symbols'))

        menu = qt.QMenu(self)
        self._addSizeSliderToMenu(menu)
        menu.addSeparator()
        self._addSymbolsToMenu(menu)

        self.setMenu(menu)
        self.setPopupMode(qt.QToolButton.InstantPopup)


class ScatterVisualizationToolButton(_SymbolToolButtonBase):
    """QToolButton to select the visualization mode of scatter plot

    :param parent: See QWidget
    :param plot: The `~silx.gui.plot.PlotWidget` to control
    """

    def __init__(self, parent=None, plot=None):
        super(ScatterVisualizationToolButton, self).__init__(
            parent=parent, plot=plot)

        self.setToolTip(
            'Set scatter visualization mode, symbol marker and size')
        self.setIcon(icons.getQIcon('eye'))

        menu = qt.QMenu(self)

        # Add visualization modes

        for mode in Scatter.supportedVisualizations():
            if mode is not Scatter.Visualization.BINNED_STATISTIC:
                name = mode.value.capitalize()
                action = qt.QAction(name, menu)
                action.setCheckable(False)
                action.triggered.connect(
                    functools.partial(self._visualizationChanged, mode, None))
                menu.addAction(action)

        if Scatter.Visualization.BINNED_STATISTIC in Scatter.supportedVisualizations():
            reductions = Scatter.supportedVisualizationParameterValues(
                Scatter.VisualizationParameter.BINNED_STATISTIC_FUNCTION)
            if reductions:
                submenu = menu.addMenu('Binned Statistic')
                for reduction in reductions:
                    name = reduction.capitalize()
                    action = qt.QAction(name, menu)
                    action.setCheckable(False)
                    action.triggered.connect(functools.partial(
                        self._visualizationChanged,
                        Scatter.Visualization.BINNED_STATISTIC,
                        {Scatter.VisualizationParameter.BINNED_STATISTIC_FUNCTION: reduction}))
                    submenu.addAction(action)

                submenu.addSeparator()
                binsmenu = submenu.addMenu('N Bins')

                slider = qt.QSlider(qt.Qt.Horizontal)
                slider.setRange(10, 1000)
                slider.setValue(100)
                slider.setTracking(False)
                slider.valueChanged.connect(self._binningChanged)
                widgetAction = qt.QWidgetAction(binsmenu)
                widgetAction.setDefaultWidget(slider)
                binsmenu.addAction(widgetAction)

        menu.addSeparator()

        submenu = menu.addMenu(icons.getQIcon('plot-symbols'), "Symbol")
        self._addSymbolsToMenu(submenu)

        submenu = menu.addMenu(icons.getQIcon('plot-symbols'), "Symbol Size")
        self._addSizeSliderToMenu(submenu)

        self.setMenu(menu)
        self.setPopupMode(qt.QToolButton.InstantPopup)

    def _visualizationChanged(self, mode, parameters=None):
        """Handle change of visualization mode.

        :param ScatterVisualizationMixIn.Visualization mode:
            The visualization mode to use for scatter
        :param Union[dict,None] parameters:
            Dict of VisualizationParameter: parameter_value to set
            with the visualization.
        """
        plot = self.plot()
        if plot is None:
            return

        for item in plot.getItems():
            if isinstance(item, Scatter):
                if parameters:
                    for parameter, value in parameters.items():
                        item.setVisualizationParameter(parameter, value)
                item.setVisualization(mode)

    def _binningChanged(self, value):
        """Handle change of binning.

        :param int value: The number of bin on each dimension.
        """
        plot = self.plot()
        if plot is None:
            return

        for item in plot.getItems():
            if isinstance(item, Scatter):
                item.setVisualizationParameter(
                    Scatter.VisualizationParameter.BINNED_STATISTIC_SHAPE,
                    (value, value))
                item.setVisualization(Scatter.Visualization.BINNED_STATISTIC)



# /*##########################################################################
#
# Copyright (c) 2021 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""This module provides the :class:`ImageDataAggregated` items of the :class:`Plot`.
"""

__authors__ = ["T. Vincent"]
__license__ = "MIT"
__date__ = "07/07/2021"

import enum
import logging
from typing import Tuple, Union

import numpy

from ....utils.enum import Enum as _Enum
from ....utils.proxy import docstring
from .axis import Axis
from .core import ItemChangedType
from .image import ImageDataBase
from ._pick import PickingResult


_logger = logging.getLogger(__name__)


class ImageDataAggregated(ImageDataBase):
    """Item displaying an image as a density map."""

    @enum.unique
    class Aggregation(_Enum):
        NONE = "none"
        "Do not aggregate data, display as is (default)"

        MAX = "max"
        "Aggregates elements with max (ignore NaNs)"

        MEAN = "mean"
        "Aggregates elements with mean (ignore NaNs)"

        MIN = "min"
        "Aggregates elements with min (ignore NaNs)"

    def __init__(self):
        super().__init__()
        self.__cacheLODData = {}
        self.__currentLOD = 0, 0
        self.__aggregationMode = self.Aggregation.NONE

    def setAggregationMode(self, mode: Union[str,Aggregation]):
        """Set the aggregation method used to reduce the data to screen resolution.

        :param Aggregation mode: The aggregation method
        """
        aggregationMode = self.Aggregation.from_value(mode)
        if aggregationMode != self.__aggregationMode:
            self.__aggregationMode = aggregationMode
            self.__cacheLODData = {}  # Clear cache
            self._updated(ItemChangedType.VISUALIZATION_MODE)

    def getAggregationMode(self) -> Aggregation:
        """Returns the currently used aggregation method."""
        return self.__aggregationMode

    def _addBackendRenderer(self, backend):
        """Update backend renderer"""
        plot = self.getPlot()
        assert plot is not None
        if not self._isPlotLinear(plot):
            # Do not render with non linear scales
            return None

        data = self.getData(copy=False)
        if data.size == 0:
            return None  # No data to display

        aggregationMode = self.getAggregationMode()
        if aggregationMode == self.Aggregation.NONE:  # Pass data as it is
            displayedData = data
            scale = self.getScale()

        else:  # Aggregate data according to level of details
            if aggregationMode == self.Aggregation.MAX:
                aggregator = numpy.nanmax
            elif aggregationMode == self.Aggregation.MEAN:
                aggregator = numpy.nanmean
            elif aggregationMode == self.Aggregation.MIN:
                aggregator = numpy.nanmin
            else:
                _logger.error("Unsupported aggregation mode")
                return None

            lodx, lody = self._getLevelOfDetails()

            if (lodx, lody) not in self.__cacheLODData:
                height, width = data.shape
                self.__cacheLODData[(lodx, lody)] = aggregator(
                    data[: (height // lody) * lody, : (width // lodx) * lodx].reshape(
                        height // lody, lody, width // lodx, lodx
                    ),
                    axis=(1, 3),
                )

            self.__currentLOD = lodx, lody
            displayedData = self.__cacheLODData[self.__currentLOD]

            sx, sy = self.getScale()
            scale = sx * lodx, sy * lody

        return backend.addImage(
            displayedData,
            origin=self.getOrigin(),
            scale=scale,
            colormap=self._getColormapForRendering(),
            alpha=self.getAlpha(),
        )

    def _getPixelSizeInData(self, axis="left"):
        """Returns the size of a pixel in plot data coordinates

        :param str axis: Y axis to use in: 'left' (default), 'right'
        :return:
            Size (width, height) of a Qt pixel in data coordinates.
            Size is None if it cannot be computed
        :rtype: Union[List[float],None]
        """
        assert axis in ("left", "right")
        plot = self.getPlot()
        if plot is None:
            return None

        xaxis = plot.getXAxis()
        yaxis = plot.getYAxis(axis)

        if (
            xaxis.getScale() != Axis.LINEAR
            or yaxis.getScale() != Axis.LINEAR
        ):
            raise RuntimeError("Only available with linear axes")

        xmin, xmax = xaxis.getLimits()
        ymin, ymax = yaxis.getLimits()
        width, height = plot.getPlotBoundsInPixels()[2:]
        if width == 0 or height == 0:
            return None
        else:
            return (xmax - xmin) / width, (ymax - ymin) / height

    def _getLevelOfDetails(self) -> Tuple[int, int]:
        """Return current level of details the image is displayed with."""
        plot = self.getPlot()
        if plot is None or not self._isPlotLinear(plot):
            return 1, 1  # Fallback to bas LOD

        sx, sy = self.getScale()
        xUnitPerPixel, yUnitPerPixel = self._getPixelSizeInData()
        lodx = max(1, int(numpy.ceil(xUnitPerPixel / sx)))
        lody = max(1, int(numpy.ceil(yUnitPerPixel / sy)))
        return lodx, lody

    @docstring(ImageDataBase)
    def setData(self, data, copy=True):
        self.__cacheLODData = {}  # Reset cache
        super().setData(data)

    @docstring(ImageDataBase)
    def _setPlot(self, plot):
        """Refresh image when plot limits change"""
        previousPlot = self.getPlot()
        if previousPlot is not None:
            for axis in (previousPlot.getXAxis(), previousPlot.getYAxis()):
                axis.sigLimitsChanged.disconnect(self.__plotLimitsChanged)

        super()._setPlot(plot)

        if plot is not None:
            for axis in (plot.getXAxis(), plot.getYAxis()):
                axis.sigLimitsChanged.connect(self.__plotLimitsChanged)

    def __plotLimitsChanged(self):
        """Trigger update if level of details has changed"""
        if (self.getAggregationMode() != self.Aggregation.NONE and
                self.__currentLOD != self._getLevelOfDetails()):
            self._updated()

    @docstring(ImageDataBase)
    def pick(self, x, y):
        result = super().pick(x, y)
        if result is None:
            return None

        # Compute indices in initial data
        plot = self.getPlot()
        if plot is None:
            return None
        dataPos = plot.pixelToData(x, y, axis="left", check=True)
        if dataPos is None:
            return None  # Outside plot area

        ox, oy = self.getOrigin()
        sx, sy = self.getScale()
        col = int((dataPos[0] - ox) / sx)
        row = int((dataPos[1] - oy) / sy)
        height, width = self.getData(copy=False).shape[:2]
        if 0 <= col < width and 0 <= row < height:
            return PickingResult(self, ((row,), (col,)))
        return None



# /*##########################################################################
#
# Copyright (c) 2016-2021 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""This module defines a widget designed to display data arrays with any
number of dimensions as 2D frames (images, slices) in a table view.
The dimensions not displayed in the table can be browsed using improved
sliders.

The widget uses a TableView that relies on a custom abstract item
model: :class:`silx.gui.data.ArrayTableModel`.
"""
import sys

from silx.gui import qt
from silx.gui.widgets.TableWidget import TableView
from .ArrayTableModel import ArrayTableModel
from silx.gui.widgets.FrameBrowser import HorizontalSliderWithBrowser

__authors__ = ["V.A. Sole", "P. Knobel"]
__license__ = "MIT"
__date__ = "24/01/2017"


class AxesSelector(qt.QWidget):
    """Widget with two combo-boxes to select two dimensions among
    all possible dimensions of an n-dimensional array.

    The first combobox contains values from :math:`0` to :math:`n-2`.

    The choices in the 2nd CB depend on the value selected in the first one.
    If the value selected in the first CB is :math:`m`, the second one lets you
    select values from :math:`m+1` to :math:`n-1`.

    The two axes can be used to select the row axis and the column axis t
    display a slice of the array data in a table view.
    """
    sigDimensionsChanged = qt.Signal(int, int)
    """Signal emitted whenever one of the comboboxes is changed.
    The signal carries the two selected dimensions."""

    def __init__(self, parent=None, n=None):
        qt.QWidget.__init__(self, parent)
        self.layout = qt.QHBoxLayout(self)
        self.layout.setContentsMargins(0, 2, 0, 2)
        self.layout.setSpacing(10)

        self.rowsCB = qt.QComboBox(self)
        self.columnsCB = qt.QComboBox(self)

        self.layout.addWidget(qt.QLabel("Rows dimension", self))
        self.layout.addWidget(self.rowsCB)
        self.layout.addWidget(qt.QLabel("    ", self))
        self.layout.addWidget(qt.QLabel("Columns dimension", self))
        self.layout.addWidget(self.columnsCB)
        self.layout.addStretch(1)

        self._slotsAreConnected = False
        if n is not None:
            self.setNDimensions(n)

    def setNDimensions(self, n):
        """Initialize combo-boxes depending on number of dimensions of array.
        Initially, the rows dimension is the second-to-last one, and the
        columns dimension is the last one.

        Link the CBs together. MAke them emit a signal when their value is
        changed.

        :param int n: Number of dimensions of array
        """
        # remember the number of dimensions and the rows dimension
        self.n = n
        self._rowsDim = n - 2

        # ensure slots are disconnected before (re)initializing widget
        if self._slotsAreConnected:
            self.rowsCB.currentIndexChanged.disconnect(self._rowDimChanged)
            self.columnsCB.currentIndexChanged.disconnect(self._colDimChanged)

        self._clear()
        self.rowsCB.addItems([str(i) for i in range(n - 1)])
        self.rowsCB.setCurrentIndex(n - 2)
        if n >= 1:
            self.columnsCB.addItem(str(n - 1))
            self.columnsCB.setCurrentIndex(0)

        # reconnect slots
        self.rowsCB.currentIndexChanged.connect(self._rowDimChanged)
        self.columnsCB.currentIndexChanged.connect(self._colDimChanged)
        self._slotsAreConnected = True

        # emit new dimensions
        if n > 2:
            self.sigDimensionsChanged.emit(n - 2, n - 1)

    def setDimensions(self, row_dim, col_dim):
        """Set the rows and columns dimensions.

        The rows dimension must be lower than the columns dimension.

        :param int row_dim: Rows dimension
        :param int col_dim: Columns dimension
        """
        if row_dim >= col_dim:
            raise IndexError("Row dimension must be lower than column dimension")
        if not (0 <= row_dim < self.n - 1):
            raise IndexError("Row dimension must be between 0 and %d" % (self.n - 2))
        if not (row_dim < col_dim <= self.n - 1):
            raise IndexError("Col dimension must be between %d and %d" % (row_dim + 1, self.n - 1))

        # set the rows dimension; this triggers an update of columnsCB
        self.rowsCB.setCurrentIndex(row_dim)
        # columnsCB first item is "row_dim + 1". So index of "col_dim" is
        # col_dim - (row_dim + 1)
        self.columnsCB.setCurrentIndex(col_dim - row_dim - 1)

    def getDimensions(self):
        """Return a 2-tuple of the rows dimension and the columns dimension.

        :return: 2-tuple of axes numbers (row_dimension, col_dimension)
        """
        return self._getRowDim(), self._getColDim()

    def _clear(self):
        """Empty the combo-boxes"""
        self.rowsCB.clear()
        self.columnsCB.clear()

    def _getRowDim(self):
        """Get rows dimension, selected in :attr:`rowsCB`
        """
        # rows combobox contains elements "0", ..."n-2",
        # so the selected dim is always equal to the index
        return self.rowsCB.currentIndex()

    def _getColDim(self):
        """Get columns dimension, selected in :attr:`columnsCB`"""
        # columns combobox contains elements "row_dim+1", "row_dim+2", ..., "n-1"
        # so the selected dim is equal to row_dim + 1 + index
        return self._rowsDim + 1 + self.columnsCB.currentIndex()

    def _rowDimChanged(self):
        """Update columns combobox when the rows dimension is changed.

        Emit :attr:`sigDimensionsChanged`"""
        old_col_dim = self._getColDim()
        new_row_dim = self._getRowDim()

        # clear cols CB
        self.columnsCB.currentIndexChanged.disconnect(self._colDimChanged)
        self.columnsCB.clear()
        # refill cols CB
        for i in range(new_row_dim + 1, self.n):
            self.columnsCB.addItem(str(i))

        # keep previous col dimension if possible
        new_col_cb_idx = old_col_dim - (new_row_dim + 1)
        if new_col_cb_idx < 0:
            # if row_dim is now greater than the previous col_dim,
            # we select a new col_dim = row_dim + 1 (first element in cols CB)
            new_col_cb_idx = 0
        self.columnsCB.setCurrentIndex(new_col_cb_idx)

        # reconnect slot
        self.columnsCB.currentIndexChanged.connect(self._colDimChanged)

        self._rowsDim = new_row_dim

        self.sigDimensionsChanged.emit(self._getRowDim(), self._getColDim())

    def _colDimChanged(self):
        """Emit :attr:`sigDimensionsChanged`"""
        self.sigDimensionsChanged.emit(self._getRowDim(), self._getColDim())


def _get_shape(array_like):
    """Return shape of an array like object.

    In case the object is a nested sequence (list of lists, tuples...),
    the size of each dimension is assumed to be uniform, and is deduced from
    the length of the first sequence.

    :param array_like: Array like object: numpy array, hdf5 dataset,
        multi-dimensional sequence
    :return: Shape of array, as a tuple of integers
    """
    if hasattr(array_like, "shape"):
        return array_like.shape

    shape = []
    subsequence = array_like
    while hasattr(subsequence, "__len__"):
        shape.append(len(subsequence))
        subsequence = subsequence[0]

    return tuple(shape)


class ArrayTableWidget(qt.QWidget):
    """This widget is designed to display data of 2D frames (images, slices)
    in a table view. The widget can load any n-dimensional array, and display
    any 2-D frame/slice in the array.

    The index of the dimensions orthogonal to the displayed frame can be set
    interactively using a browser widget (sliders, buttons and text entries).

    To set the data, use :meth:`setArrayData`.
    To select the perspective, use :meth:`setPerspective` or
    use :meth:`setFrameAxes`.
    To select the frame, use :meth:`setFrameIndex`.

    .. image:: img/ArrayTableWidget.png
    """
    def __init__(self, parent=None):
        """

        :param parent: parent QWidget
        :param labels: list of labels for each dimension of the array
        """
        qt.QWidget.__init__(self, parent)
        self.mainLayout = qt.QVBoxLayout(self)
        self.mainLayout.setContentsMargins(0, 0, 0, 0)
        self.mainLayout.setSpacing(0)

        self.browserContainer = qt.QWidget(self)
        self.browserLayout = qt.QGridLayout(self.browserContainer)
        self.browserLayout.setContentsMargins(0, 0, 0, 0)
        self.browserLayout.setSpacing(0)

        self._dimensionLabelsText = []
        """List of text labels sorted in the increasing order of the dimension
        they apply to."""
        self._browserLabels = []
        """List of QLabel widgets."""
        self._browserWidgets = []
        """List of HorizontalSliderWithBrowser widgets."""

        self.axesSelector = AxesSelector(self)

        self.view = TableView(self)

        self.mainLayout.addWidget(self.browserContainer)
        self.mainLayout.addWidget(self.axesSelector)
        self.mainLayout.addWidget(self.view)

        self.model = ArrayTableModel(self)
        self.view.setModel(self.model)

    def setArrayData(self, data, labels=None, copy=True, editable=False):
        """Set the data array. Update frame browsers and labels.

        :param data: Numpy array or similar object (e.g. nested sequence,
            h5py dataset...)
        :param labels: list of labels for each dimension of the array, or
            boolean ``True`` to use default labels ("dimension 0",
            "dimension 1", ...). `None` to disable labels (default).
        :param bool copy: If *True*, store a copy of *data* in the model. If
            *False*, store a reference to *data* if possible (only possible if
            *data* is a proper numpy array or an object that implements the
            same methods).
        :param bool editable: Flag to enable editing data. Default is *False*
        """
        self._data_shape = _get_shape(data)

        n_widgets = len(self._browserWidgets)
        n_dimensions = len(self._data_shape)

        # Reset text of labels
        self._dimensionLabelsText = []
        for i in range(n_dimensions):
            if labels in [True, 1]:
                label_text = "Dimension %d" % i
            elif labels is None or i >= len(labels):
                label_text = ""
            else:
                label_text = labels[i]
            self._dimensionLabelsText.append(label_text)

        # not enough widgets, create new ones (we need n_dim - 2)
        for i in range(n_widgets, n_dimensions - 2):
            browser = HorizontalSliderWithBrowser(self.browserContainer)
            self.browserLayout.addWidget(browser, i, 1)
            self._browserWidgets.append(browser)
            browser.valueChanged.connect(self._browserSlot)
            browser.setEnabled(False)
            browser.hide()

            label = qt.QLabel(self.browserContainer)
            self._browserLabels.append(label)
            self.browserLayout.addWidget(label, i, 0)
            label.hide()

        n_widgets = len(self._browserWidgets)
        for i in range(n_widgets):
            label = self._browserLabels[i]
            browser = self._browserWidgets[i]

            if (i + 2) < n_dimensions:
                label.setText(self._dimensionLabelsText[i])
                browser.setRange(0, self._data_shape[i] - 1)
                browser.setEnabled(True)
                browser.show()
                if labels is not None:
                    label.show()
                else:
                    label.hide()
            else:
                browser.setEnabled(False)
                browser.hide()
                label.hide()

        # set model
        self.model.setArrayData(data, copy=copy, editable=editable)
        # some linux distributions need this call
        self.view.setModel(self.model)
        if editable:
            self.view.enableCut()
            self.view.enablePaste()

        # initialize & connect axesSelector
        self.axesSelector.setNDimensions(n_dimensions)
        self.axesSelector.sigDimensionsChanged.connect(self.setFrameAxes)

    def setArrayColors(self, bgcolors=None, fgcolors=None):
        """Set the colors for all table cells by passing an array
        of RGB or RGBA values (integers between 0 and 255).

        The shape of the colors array must be consistent with the data shape.

        If the data array is n-dimensional, the colors array must be
        (n+1)-dimensional, with the first n-dimensions identical to the data
        array dimensions, and the last dimension length-3 (RGB) or
        length-4 (RGBA).

        :param bgcolors: RGB or RGBA colors array, defining the background color
            for each cell in the table.
        :param fgcolors: RGB or RGBA colors array, defining the foreground color
            (text color) for each cell in the table.
        """
        self.model.setArrayColors(bgcolors, fgcolors)

    def displayAxesSelector(self, isVisible):
        """Allow to display or hide the axes selector.

        :param bool isVisible: True to display the axes selector.
        """
        self.axesSelector.setVisible(isVisible)

    def setFrameIndex(self, index):
        """Set the active slice/image index in the n-dimensional array.

        A frame is a 2D array extracted from an array. This frame is
        necessarily parallel to 2 axes, and orthogonal to all other axes.

        The index of a frame is a sequence of indices along the orthogonal
        axes, where the frame intersects the respective axis. The indices
        are listed in the same order as the corresponding dimensions of the
        data array.

        For example, it the data array has 5 dimensions, and we are
        considering frames whose parallel axes are the 2nd and 4th dimensions
        of the array, the frame index will be a sequence of length 3
        corresponding to the indices where the frame intersects the 1st, 3rd
        and 5th axes.

        :param index: Sequence of indices defining the active data slice in
            a n-dimensional array. The sequence length is :math:`n-2`
        :raise: IndexError if any index in the index sequence is out of bound
            on its respective axis.
        """
        self.model.setFrameIndex(index)

    def _resetBrowsers(self, perspective):
        """Adjust limits for browsers based on the perspective and the
        size of the corresponding dimensions. Reset the index to 0.
        Update the dimension in the labels.

        :param perspective: Sequence of axes/dimensions numbers (0-based)
            defining the axes orthogonal to the frame.
        """
        # for 3D arrays we can accept an int rather than a 1-tuple
        if not hasattr(perspective, "__len__"):
            perspective = [perspective]

        # perspective must be sorted
        perspective = sorted(perspective)

        n_dimensions = len(self._data_shape)
        for i in range(n_dimensions - 2):
            browser = self._browserWidgets[i]
            label = self._browserLabels[i]
            browser.setRange(0, self._data_shape[perspective[i]] - 1)
            browser.setValue(0)
            label.setText(self._dimensionLabelsText[perspective[i]])

    def setPerspective(self, perspective):
        """Set the *perspective* by specifying which axes are orthogonal
        to the frame.

        For the opposite approach (defining parallel axes), use
        :meth:`setFrameAxes` instead.

        :param perspective: Sequence of unique axes numbers (0-based) defining
            the orthogonal axes. For a n-dimensional array, the sequence
            length is :math:`n-2`. The order is of the sequence is not taken
            into account (the dimensions are displayed in increasing order
            in the widget).
        """
        self.model.setPerspective(perspective)
        self._resetBrowsers(perspective)

    def setFrameAxes(self, row_axis, col_axis):
        """Set the *perspective* by specifying which axes are parallel
        to the frame.

        For the opposite approach (defining orthogonal axes), use
        :meth:`setPerspective` instead.

        :param int row_axis: Index (0-based) of the first dimension used as a frame
            axis
        :param int col_axis: Index (0-based) of the 2nd dimension used as a frame
            axis
        """
        self.model.setFrameAxes(row_axis, col_axis)
        n_dimensions = len(self._data_shape)
        perspective = tuple(set(range(0, n_dimensions)) - {row_axis, col_axis})
        self._resetBrowsers(perspective)

    def _browserSlot(self, value):
        index = []
        for browser in self._browserWidgets:
            if browser.isEnabled():
                index.append(browser.value())
        self.setFrameIndex(index)
        self.view.reset()

    def getData(self, copy=True):
        """Return a copy of the data array, or a reference to it if
        *copy=False* is passed as parameter.

        :param bool copy: If *True* (default), return a copy of the data. If
            *False*, return a reference.
        :return: Numpy array of data, or reference to original data object
            if *copy=False*
        """
        return self.model.getData(copy=copy)


def main():
    import numpy
    a = qt.QApplication([])
    d = numpy.random.normal(0, 1, (4, 5, 1000, 1000))
    for j in range(4):
        for i in range(5):
            d[j, i, :, :] += i + 10 * j
    w = ArrayTableWidget()
    if "2" in sys.argv:
        print("sending a single image")
        w.setArrayData(d[0, 0])
    elif "3" in sys.argv:
        print("sending 5 images")
        w.setArrayData(d[0])
    else:
        print("sending 4 * 5 images ")
        w.setArrayData(d, labels=True)
    w.show()
    a.exec()

if __name__ == "__main__":
    main()



# /*##########################################################################
#
# Copyright (c) 2016-2022 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

"""This module initializes matplotlib and sets-up the backend to use.

It MUST be imported prior to any other import of matplotlib.

It provides the matplotlib :class:`FigureCanvasQTAgg` class corresponding
to the used backend.
"""

__authors__ = ["T. Vincent"]
__license__ = "MIT"
__date__ = "02/05/2018"


import io
from pkg_resources import parse_version
import matplotlib
import numpy

from .. import qt


def rasterMathText(text, font, size=-1, weight=-1, italic=False, devicePixelRatio=1.0):
    """Raster text using matplotlib supporting latex-like math syntax.

    It supports multiple lines.

    :param str text: The text to raster
    :param font: Font name or QFont to use
    :type font: str or :class:`QFont`
    :param int size:
        Font size in points
        Used only if font is given as name.
    :param int weight:
        Font weight in [0, 99], see QFont.Weight.
        Used only if font is given as name.
    :param bool italic:
        True for italic font (default: False).
        Used only if font is given as name.
    :param float devicePixelRatio:
        The current ratio between device and device-independent pixel
        (default: 1.0)
    :return: Corresponding image in gray scale and baseline offset from top
    :rtype: (HxW numpy.ndarray of uint8, int)
    """
    # Implementation adapted from:
    # https://github.com/matplotlib/matplotlib/blob/d624571a19aec7c7d4a24123643288fc27db17e7/lib/matplotlib/mathtext.py#L264

    # Lazy import to avoid imports before setting matplotlib's rcParams
    from matplotlib.font_manager import FontProperties
    from matplotlib.mathtext import MathTextParser
    from matplotlib import figure

    dpi = 96  # default
    qapp = qt.QApplication.instance()
    if qapp:
        screen = qapp.primaryScreen()
        if screen:
            dpi = screen.logicalDotsPerInchY()

    # Make sure dpi is even, it causes issues with array reshape otherwise
    dpi = ((dpi * devicePixelRatio) // 2) * 2

    stripped_text = text.strip("\n")

    parser = MathTextParser("path")
    width, height, depth, _, _ = parser.parse(stripped_text, dpi=dpi)
    width *= 2
    height *= 2 * (stripped_text.count("\n") + 1)

    if not isinstance(font, qt.QFont):
        font = qt.QFont(font, size, weight, italic)
    prop = FontProperties(
        family=font.family(),
        style="italic" if font.italic() else "normal",
        weight=10 * font.weight(),
        size=font.pointSize(),
    )

    fig = figure.Figure(figsize=(width / dpi, height / dpi))
    fig.text(0, depth / height, stripped_text, fontproperties=prop)
    with io.BytesIO() as buffer:
        fig.savefig(buffer, dpi=dpi, format="raw")
        buffer.seek(0)
        image = numpy.frombuffer(buffer.read(), dtype=numpy.uint8).reshape(
            int(height), int(width), 4
        )

    # RGB to inverted R channel
    array = 255 - image[:, :, 0]

    # Remove leading and trailing empty columns/rows but one on each side
    filled_rows = numpy.nonzero(numpy.sum(array, axis=1))[0]
    filled_columns = numpy.nonzero(numpy.sum(array, axis=0))[0]
    if len(filled_rows) == 0 or len(filled_columns) == 0:
        return array, image.shape[0] - 1

    clipped_array = numpy.ascontiguousarray(
        array[
            max(0, filled_rows[0] - 1) : filled_rows[-1] + 2,
            max(0, filled_columns[0] - 1) : filled_columns[-1] + 2,
        ]
    )

    return clipped_array, image.shape[0] - 1  # baseline not available


def _matplotlib_use(backend, force):
    """Wrapper of `matplotlib.use` to set-up backend.

    It adds extra initialization for PySide2 with matplotlib < 2.2.
    """
    # This is kept for compatibility with matplotlib < 2.2
    if (
        parse_version(matplotlib.__version__) < parse_version("2.2")
        and qt.BINDING == "PySide2"
    ):
        matplotlib.rcParams["backend.qt5"] = "PySide2"

    matplotlib.use(backend, force=force)


if qt.BINDING in ("PySide6", "PyQt6", "PyQt5", "PySide2"):
    _matplotlib_use("Qt5Agg", force=False)
    from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg  # noqa

else:
    raise ImportError("Unsupported Qt binding: %s" % qt.BINDING)



# /*##########################################################################
#
# Copyright (c) 2017-2019 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""This module provides a QMainWindow with a 3D SceneWidget and toolbars.
"""

__authors__ = ["T. Vincent"]
__license__ = "MIT"
__date__ = "29/11/2017"


from ...gui import qt, icons
from ...gui.widgets.BoxLayoutDockWidget import BoxLayoutDockWidget

from .actions.mode import InteractiveModeAction
from .SceneWidget import SceneWidget
from .tools import OutputToolBar, InteractiveModeToolBar, ViewpointToolBar
from .tools.GroupPropertiesWidget import GroupPropertiesWidget
from .tools.PositionInfoWidget import PositionInfoWidget

from .ParamTreeView import ParamTreeView

# Imported here for convenience
from . import items  # noqa


__all__ = ['items', 'SceneWidget', 'SceneWindow']


class _PanPlaneAction(InteractiveModeAction):
    """QAction to set plane pan interaction on a Plot3DWidget

    :param parent: See :class:`QAction`
    :param ~silx.gui.plot3d.Plot3DWidget.Plot3DWidget plot3d:
        Plot3DWidget the action is associated with
    """
    def __init__(self, parent, plot3d=None):
        super(_PanPlaneAction, self).__init__(
            parent, 'panSelectedPlane', plot3d)
        self.setIcon(icons.getQIcon('3d-plane-pan'))
        self.setText('Pan plane')
        self.setCheckable(True)
        self.setToolTip(
            'Pan selected plane. Press <b>Ctrl</b> to rotate the scene.')

    def _planeChanged(self, event):
        """Handle plane updates"""
        if event in (items.ItemChangedType.VISIBLE,
                     items.ItemChangedType.POSITION):
            plane = self.sender()

            isPlaneInteractive = \
                plane._getPlane().plane.isPlane and plane.isVisible()

            if isPlaneInteractive != self.isEnabled():
                self.setEnabled(isPlaneInteractive)
                mode = 'panSelectedPlane' if isPlaneInteractive else 'rotate'
                self.getPlot3DWidget().setInteractiveMode(mode)

    def _selectionChanged(self, current, previous):
        """Handle selected object change"""
        if isinstance(previous, items.PlaneMixIn):
            previous.sigItemChanged.disconnect(self._planeChanged)

        if isinstance(current, items.PlaneMixIn):
            current.sigItemChanged.connect(self._planeChanged)
            self.setEnabled(True)
            self.getPlot3DWidget().setInteractiveMode('panSelectedPlane')
        else:
            self.setEnabled(False)

    def setPlot3DWidget(self, widget):
        previous = self.getPlot3DWidget()
        if isinstance(previous, SceneWidget):
            previous.selection().sigCurrentChanged.disconnect(
                self._selectionChanged)
            self._selectionChanged(
                None, previous.selection().getCurrentItem())

        super(_PanPlaneAction, self).setPlot3DWidget(widget)

        if isinstance(widget, SceneWidget):
            self._selectionChanged(widget.selection().getCurrentItem(), None)
            widget.selection().sigCurrentChanged.connect(
                self._selectionChanged)


class SceneWindow(qt.QMainWindow):
    """OpenGL 3D scene widget with toolbars."""

    def __init__(self, parent=None):
        super(SceneWindow, self).__init__(parent)
        if parent is not None:
            # behave as a widget
            self.setWindowFlags(qt.Qt.Widget)

        self._sceneWidget = SceneWidget()
        self.setCentralWidget(self._sceneWidget)

        # Add PositionInfoWidget to display picking info
        self._positionInfo = PositionInfoWidget()
        self._positionInfo.setSceneWidget(self._sceneWidget)

        dock = BoxLayoutDockWidget()
        dock.setWindowTitle("Selection Info")
        dock.setWidget(self._positionInfo)
        self.addDockWidget(qt.Qt.BottomDockWidgetArea, dock)

        self._interactiveModeToolBar = InteractiveModeToolBar(parent=self)
        panPlaneAction = _PanPlaneAction(self, plot3d=self._sceneWidget)
        self._interactiveModeToolBar.addAction(
            self._positionInfo.toggleAction())
        self._interactiveModeToolBar.addAction(panPlaneAction)

        self._viewpointToolBar = ViewpointToolBar(parent=self)
        self._outputToolBar = OutputToolBar(parent=self)

        for toolbar in (self._interactiveModeToolBar,
                        self._viewpointToolBar,
                        self._outputToolBar):
            toolbar.setPlot3DWidget(self._sceneWidget)
            self.addToolBar(toolbar)
            self.addActions(toolbar.actions())

        self._paramTreeView = ParamTreeView()
        self._paramTreeView.setModel(self._sceneWidget.model())

        selectionModel = self._paramTreeView.selectionModel()
        self._sceneWidget.selection()._setSyncSelectionModel(
            selectionModel)

        paramDock = qt.QDockWidget()
        paramDock.setWindowTitle('Object parameters')
        paramDock.setWidget(self._paramTreeView)
        self.addDockWidget(qt.Qt.RightDockWidgetArea, paramDock)

        self._sceneGroupResetWidget = GroupPropertiesWidget()
        self._sceneGroupResetWidget.setGroup(
            self._sceneWidget.getSceneGroup())

        resetDock = qt.QDockWidget()
        resetDock.setWindowTitle('Global parameters')
        resetDock.setWidget(self._sceneGroupResetWidget)
        self.addDockWidget(qt.Qt.RightDockWidgetArea, resetDock)
        self.tabifyDockWidget(paramDock, resetDock)

        paramDock.raise_()

    def getSceneWidget(self):
        """Returns the SceneWidget of this window.

        :rtype: ~silx.gui.plot3d.SceneWidget.SceneWidget
        """
        return self._sceneWidget

    def getGroupResetWidget(self):
        """Returns the :class:`GroupPropertiesWidget` of this window.

        :rtype: GroupPropertiesWidget
        """
        return self._sceneGroupResetWidget

    def getParamTreeView(self):
        """Returns the :class:`ParamTreeView` of this window.

        :rtype: ParamTreeView
        """
        return self._paramTreeView

    def getInteractiveModeToolBar(self):
        """Returns the interactive mode toolbar.

        :rtype: ~silx.gui.plot3d.tools.InteractiveModeToolBar
        """
        return self._interactiveModeToolBar

    def getViewpointToolBar(self):
        """Returns the viewpoint toolbar.

        :rtype: ~silx.gui.plot3d.tools.ViewpointToolBar
        """
        return self._viewpointToolBar

    def getOutputToolBar(self):
        """Returns the output toolbar.

        :rtype: ~silx.gui.plot3d.tools.OutputToolBar
        """
        return self._outputToolBar

    def getPositionInfoWidget(self):
        """Returns the widget displaying selected position information.

        :rtype: ~silx.gui.plot3d.tools.PositionInfoWidget.PositionInfoWidget
        """
        return self._positionInfo



# /*##########################################################################
# Copyright (C) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ############################################################################*/
"""
This module provides utility methods for retrying methods until they
no longer fail.
"""

__authors__ = ["W. de Nolf"]
__license__ = "MIT"
__date__ = "05/02/2020"


import time
import inspect
from functools import wraps
from contextlib import contextmanager
import multiprocessing
from queue import Empty


RETRY_PERIOD = 0.01


class RetryTimeoutError(TimeoutError):
    pass


class RetryError(RuntimeError):
    pass


def _default_retry_on_error(e):
    """
    :param BaseException e:
    :returns bool:
    """
    return isinstance(e, RetryError)


@contextmanager
def _handle_exception(retry_state):
    try:
        yield
    except BaseException as e:
        retry_on_error = retry_state.get("retry_on_error")
        if retry_on_error is not None and retry_on_error(e):
            retry_state["exception"] = e
        else:
            raise


def _retry_loop(retry_timeout=None, retry_period=None, retry_on_error=None):
    """Iterator which is endless or ends with an RetryTimeoutError.
    It yields a dictionary which can be used to influence the loop.

    :param num retry_timeout:
    :param num retry_period: sleep before retry
    :param callable or None retry_on_error: checks whether an exception is
                                            eligible for retry
    """
    has_timeout = retry_timeout is not None
    if has_timeout:
        t0 = time.time()
    else:
        t0 = None
    retry_state = {"t0": t0, "exception": None, "retry_on_error": retry_on_error}
    while True:
        yield retry_state
        if retry_period is not None:
            time.sleep(retry_period)
        if has_timeout and (time.time() - t0) > retry_timeout:
            err_msg = "%s seconds" % retry_timeout
            cause = retry_state.get("exception")
            raise RetryTimeoutError(err_msg) from cause


def retry(
    retry_timeout=None, retry_period=None, retry_on_error=_default_retry_on_error
):
    """Decorator for a method that needs to be executed until it not longer
    fails or until `retry_on_error` returns False.

    The decorator arguments can be overriden by using them when calling the
    decorated method.

    Generator functions are required to have a `start_index` argument which allows
    the method to start iterating from the last failure when called on retry.

    :param num retry_timeout:
    :param num retry_period: sleep before retry
    :param callable or None retry_on_error: checks whether an exception is
                                            eligible for retry
    """

    if retry_period is None:
        retry_period = RETRY_PERIOD

    def decorator(method):
        if inspect.isgeneratorfunction(method):
            if "start_index" not in inspect.signature(method).parameters:
                raise TypeError(
                    "The generator function '%s' needs a `start_index` named argument because it is wrapped with the `retry` decorator."
                    % method.__name__
                )

            @wraps(method)
            def wrapper(*args, **kw):
                _retry_timeout = kw.pop("retry_timeout", retry_timeout)
                _retry_period = kw.pop("retry_period", retry_period)
                _retry_on_error = kw.pop("retry_on_error", retry_on_error)
                start_index = kw.pop("start_index", 0)
                if start_index is None:
                    start_index = 0
                for retry_state in _retry_loop(
                    retry_timeout=_retry_timeout,
                    retry_period=_retry_period,
                    retry_on_error=_retry_on_error,
                ):
                    with _handle_exception(retry_state):
                        oretry_on_error = retry_state["retry_on_error"]
                        for result in method(*args, start_index=start_index, **kw):
                            start_index += 1
                            retry_state["retry_on_error"] = None
                            # any exception here will NOT cause a retry
                            yield result
                            # restart the retry loop
                            if retry_state["t0"] is not None:
                                retry_state["t0"] = time.time()
                            retry_state["retry_on_error"] = oretry_on_error
                        return

        else:

            @wraps(method)
            def wrapper(*args, **kw):
                _retry_timeout = kw.pop("retry_timeout", retry_timeout)
                _retry_period = kw.pop("retry_period", retry_period)
                _retry_on_error = kw.pop("retry_on_error", retry_on_error)
                for retry_state in _retry_loop(
                    retry_timeout=_retry_timeout,
                    retry_period=_retry_period,
                    retry_on_error=_retry_on_error,
                ):
                    with _handle_exception(retry_state):
                        return method(*args, **kw)

        return wrapper

    return decorator


def retry_contextmanager(
    retry_timeout=None, retry_period=None, retry_on_error=_default_retry_on_error
):
    """Decorator to make a context manager from a method that needs to be
    entered until it no longer fails or until `retry_on_error` returns False.

    The decorator arguments can be overriden by using them when calling the
    decorated method.

    :param num retry_timeout:
    :param num retry_period: sleep before retry
    :param callable or None retry_on_error: checks whether an exception is
                                            eligible for retry
    """

    if retry_period is None:
        retry_period = RETRY_PERIOD

    def decorator(method):
        @wraps(method)
        def wrapper(*args, **kw):
            _retry_timeout = kw.pop("retry_timeout", retry_timeout)
            _retry_period = kw.pop("retry_period", retry_period)
            _retry_on_error = kw.pop("retry_on_error", retry_on_error)
            for retry_state in _retry_loop(
                retry_timeout=_retry_timeout,
                retry_period=_retry_period,
                retry_on_error=_retry_on_error,
            ):
                with _handle_exception(retry_state):
                    ctx = method(*args, **kw)
                    result = next(ctx)
                    retry_state["retry_on_error"] = None
                    yield result
                    try:
                        next(ctx)
                    except StopIteration:
                        return
                    else:
                        raise RuntimeError(str(method) + " should only yield once")

        return contextmanager(wrapper)

    return decorator


def _subprocess_main(queue, method, retry_on_error, *args, **kw):
    try:
        result = method(*args, **kw)
    except BaseException as e:
        if retry_on_error(e):
            # As the traceback gets lost, make sure the top-level
            # exception is RetryError
            e = RetryError(str(e))
        queue.put(e)
    else:
        queue.put(result)


def retry_in_subprocess(
    retry_timeout=None, retry_period=None, retry_on_error=_default_retry_on_error
):
    """Same as `retry` but it also retries segmentation faults.

    As subprocesses are spawned, you cannot use this decorator with the "@" syntax
    because the decorated method needs to be an attribute of a module:

    .. code-block:: python

        def _method(*args, **kw):
            ...

        method = retry_in_subprocess()(_method)

    :param num retry_timeout:
    :param num retry_period: sleep before retry
    :param callable or None retry_on_error: checks whether an exception is
                                            eligible for retry
    """

    if retry_period is None:
        retry_period = RETRY_PERIOD

    def decorator(method):
        @wraps(method)
        def wrapper(*args, **kw):
            _retry_timeout = kw.pop("retry_timeout", retry_timeout)
            _retry_period = kw.pop("retry_period", retry_period)
            _retry_on_error = kw.pop("retry_on_error", retry_on_error)

            ctx = multiprocessing.get_context("spawn")

            def start_subprocess():
                queue = ctx.Queue(maxsize=1)
                p = ctx.Process(
                    target=_subprocess_main,
                    args=(queue, method, retry_on_error) + args,
                    kwargs=kw,
                )
                p.start()
                return p, queue

            def stop_subprocess(p):
                try:
                    p.kill()
                except AttributeError:
                    p.terminate()
                p.join()

            p, queue = start_subprocess()
            try:
                for retry_state in _retry_loop(
                    retry_timeout=_retry_timeout, retry_on_error=_retry_on_error
                ):
                    with _handle_exception(retry_state):
                        if not p.is_alive():
                            p, queue = start_subprocess()
                        try:
                            result = queue.get(block=True, timeout=_retry_period)
                        except Empty:
                            pass
                        except ValueError:
                            pass
                        else:
                            if isinstance(result, BaseException):
                                stop_subprocess(p)
                                raise result
                            else:
                                return result
            finally:
                stop_subprocess(p)

        return wrapper

    return decorator
